import os
import numpy as np
import casadi as ca
import forwardkinematics.urdfFks.casadiConversion.urdfparser as u2c

from forwardkinematics.fksCommon.fk import ForwardKinematics


class LinkNotInURDFError(Exception):
    pass


class URDFForwardKinematics(ForwardKinematics):
    def __init__(self, fileName, links, rootLink, end_links, n):
        super().__init__()
        self._urdf_file = (
            os.path.dirname(os.path.abspath(__file__)) + "/urdf/" + fileName
        )
        self._links = links
        self._rootLink = rootLink
        self._end_links = end_links
        self._q_ca = ca.SX.sym("q", n)
        self.readURDF()
        self._n = n
        self.generateFunctions()

    def readURDF(self):
        self.robot = u2c.URDFparser(rootLink=self._rootLink, end_links=self._end_links)
        self.robot.from_file(self._urdf_file)
        self.robot.detect_link_names()
        self.robot.set_joint_variable_map()

    def generateFunctions(self):
        self._fks = {}
        for link in self.robot.link_names():
            ca_fun = ca.Function(
                "fk" + link, [self._q_ca], [self.casadi_by_name(self._q_ca, link)]
            )
            self._fks[link] = ca_fun

    def fk(self, q, link: str, positionOnly: bool = False):
        if isinstance(link, str):
            return self.fk_by_name(q, link, positionOnly=positionOnly)
        else:
            return super().fk(q, link, positionOnly=positionOnly)

    def fk_by_name(self, q: ca.SX, link: str, positionOnly=False):
        if isinstance(q, ca.SX):
            return self.casadi_by_name(q, link, positionOnly=positionOnly)
        elif isinstance(q, np.ndarray):
            return self.numpy_by_name(q, link, positionOnly=positionOnly)

    def casadi(self, q: ca.SX, i: int, positionOnly=False):
        return self.casadi_by_name(q, self._links[i], positionOnly=positionOnly)

    def casadi_by_name(self, q: ca.SX, link: str, positionOnly=False):
        """
        Raises:
            LinkNotInURDFError: An error occured accessing the urdf link.
        """
        if link not in self.robot.link_names():
            raise LinkNotInURDFError(
                f"""The link you have requested, {link}, is not in the urdf.
                    Possible links are  {self.robot.link_names()}"""
            )
        if positionOnly:
            return self.robot.get_forward_kinematics(self._rootLink, link, q)["T_fk"][
                0:3, 3
            ]
        else:
            return self.robot.get_forward_kinematics(self._rootLink, link, q)["T_fk"]

    def numpy(self, q: ca.SX, i: int, positionOnly=False):
        return self.numpy_by_name(q, self._links[i], positionOnly=positionOnly)

    def numpy_by_name(self, q: np.ndarray, link: str, positionOnly=False):
        """
        Raises:
            LinkNotInURDFError: An error occured accessing the urdf link.
        """
        if link not in self.robot.link_names():
            raise LinkNotInURDFError(
                f"""The link you have requested, {link}, is not in the urdf.
                    Possible links are  {self.robot.link_names()}"""
            )
        if positionOnly:
            return np.array(self._fks[link](q))[0:3, 3]
        else:
            return np.array(self._fks[link](q))



import casadi as ca


class FkCreator(object):
    def __init__(self, robotType, n=3):
        if robotType == 'panda':
            from forwardkinematics.urdfFks.pandaFk import PandaFk
            self._fk = PandaFk()
        elif robotType == 'boxer':
            from forwardkinematics.urdfFks.boxerFk import BoxerFk
            self._fk = BoxerFk()
        elif robotType == 'albert':
            from forwardkinematics.urdfFks.albertFk import AlbertFk
            self._fk = AlbertFk()
        elif robotType == 'tiago':
            from forwardkinematics.urdfFks.tiagoFk import TiagoFk
            self._fk = TiagoFk()
        elif robotType == 'planarArm':
            from forwardkinematics.planarFks.planarArmFk import PlanarArmFk
            self._fk = PlanarArmFk(n)
        elif robotType == 'mobilePanda':
            from forwardkinematics.urdfFks.mobilePandaFk import MobilePandaFk
            self._fk = MobilePandaFk()
        elif robotType == 'pointRobot' or robotType == 'pointMass':
            from forwardkinematics.planarFks.pointFk import PointFk
            self._fk = PointFk()
        elif robotType == 'groundRobot':
            from forwardkinematics.planarFks.groundRobotFk import GroundRobotFk
            self._fk = GroundRobotFk()
        elif robotType == 'pointRobotUrdf':
            from forwardkinematics.urdfFks.pointRobotUrdfFk import PointRobotUrdfFk
            self._fk = PointRobotUrdfFk()
        elif robotType == 'dualArm':
            from forwardkinematics.urdfFks.dual_arm_fk import DualArmFk
            self._fk = DualArmFk()

    def fk(self):
        return self._fk



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class PandaFk(URDFForwardKinematics):
    def __init__(self):
        fileName = "panda.urdf"
        relevantLinks = ["panda_link" + str(i) for i in [0, 3, 4, 5, 6, 7, 8, 9]]
        super().__init__(fileName, relevantLinks, "panda_link0", "panda_link9", 7)



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class BoxerFk(URDFForwardKinematics):
    def __init__(self):
        fileName = "boxer.urdf"
        relevantLinks = ["origin", "ee_link", "ee_link", "ee_link"]
        super().__init__(fileName, relevantLinks, "world", "ee_link", 3)



import numpy as np
import casadi as ca
from forwardkinematics.planarFks.planar_fk import ForwardKinematicsPlanar


class PlanarArmFk(ForwardKinematicsPlanar):
    def __init__(self, n):
        super().__init__()
        self._n = n

    def casadi(self, q, i, positionOnly=False, endlink=0.0):
        fk = ca.vertcat(np.zeros(2), q[0])
        for i in range(1, i + 1):
            fk[0] += ca.cos(fk[2]) * 1.0
            fk[1] += ca.sin(fk[2]) * 1.0
            if i < q.size(1):
                fk[2] += q[i]
        fk[0] += ca.cos(fk[2]) * endlink
        fk[1] += ca.sin(fk[2]) * endlink
        if positionOnly:
            return fk[0:2]
        else:
            return fk

    def numpy(self, q, i, positionOnly=False, endlink=0.0):
        fk = np.array([0.0, 0.0, q[0]])
        for i in range(1, i + 1):
            fk[0] += np.cos(fk[2]) * 1.0
            fk[1] += np.sin(fk[2]) * 1.0
            if i < len(q):
                fk[2] += q[i]
        fk[0] += np.cos(fk[2]) * endlink
        fk[1] += np.sin(fk[2]) * endlink
        if positionOnly:
            return fk[0:2]
        else:
            return fk



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class PointRobotUrdfFk(URDFForwardKinematics):
    def __init__(self):
        fileName = "pointRobot.urdf"
        relevantLinks = ["origin", "base_link", "base_link", "base_link"]
        super().__init__(fileName, relevantLinks, "world", "base_link", 3)



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class AlbertFk(URDFForwardKinematics):
    def __init__(self):
        fileName = "albert.urdf"
        relevantLinks = ["origin", "base_tip_link", "top_mount"] + [
            "panda_link" + str(i) for i in [0, 3, 4, 5, 6, 7, 8, 9]
        ]
        super().__init__(fileName, relevantLinks, "world", ["panda_link9", "base_tip_link"], len(relevantLinks) - 1)



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class MobilePandaFk(URDFForwardKinematics):
    def __init__(self):
        fileName = "mobilePanda.urdf"
        relevantLinks = ["world", "base_link_y", "base_link"] + [
            "panda_link" + str(i) for i in [0, 3, 4, 5, 6, 7, 8, 9]
        ]
        super().__init__(fileName, relevantLinks, "world", "panda_link9", 10)



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class TiagoFk(URDFForwardKinematics):
    def __init__(self):
        fileName = "tiago.urdf"
        linkIndices = [1, 2, 3, 4, 5, 7]
        relevantLinks = ["origin", "base_link", "torso_fixed_link", "torso_fixed_link", "torso_lift_link"]
        relevantLinks += ["arm_left_" + str(i) + "_link" for i in linkIndices] + [
            "gripper_left_grasping_frame"
        ]
        relevantLinks += ["arm_right_" + str(i) + "_link" for i in linkIndices] + [
            "gripper_right_grasping_frame"
        ]
        relevantLinks += ["head_1_link", "head_2_link"]
        super().__init__(fileName, relevantLinks, "world", ["gripper_left_grasping_frame", "gripper_right_grasping_frame", "head_2_link"], len(relevantLinks) - 1)



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics, LinkNotInURDFError
import forwardkinematics.urdfFks.casadiConversion.urdfparser as u2c

class GenericURDFFk(URDFForwardKinematics):
    def __init__(self, fileName, rootLink='base_link', end_link=None):
        self._urdf_file = fileName
        self._links = []
        self._end_link = end_link
        self._rootLink = rootLink
        self.readURDF(rootLink, end_link)
        self._n = self.robot.degrees_of_freedom()
        self._q_ca = ca.SX.sym("q", self._n)
        self.generateFunctions()

    def readURDF(self, rootLink: str, end_link: str):
        self.robot = u2c.URDFparser(rootLink, end_link)
        self.robot.from_string(self._urdf_file)
        self.robot.detect_link_names()
        self.robot.set_joint_variable_map()


    def casadi(self, q: ca.SX, parent_link: str, child_link: str, positionOnly=False):
        if parent_link not in self.robot.link_names():
            raise LinkNotInURDFError(
                f"""The link you have requested, {parent_link}, is not in the urdf.
                    Possible links are  {self.robot.link_names()}"""
            )
        if child_link not in self.robot.link_names():
            raise LinkNotInURDFError(
                f"""The link you have requested, {child_link}, is not in the urdf.
                    Possible links are  {self.robot.link_names()}"""
            )
        if positionOnly:
            return self.robot.get_forward_kinematics(parent_link, child_link, q)["T_fk"][
                0:3, 3
            ]
        else:
            return self.robot.get_forward_kinematics(parent_link, child_link, q)["T_fk"]

    def fk(self, q: ca.SX, parent_link: str, child_link: str, positionOnly=False):
        if isinstance(q, ca.SX):
            return self.casadi(q, parent_link, child_link, positionOnly=positionOnly)
        elif isinstance(q, np.ndarray):
            return self.numpy(q, parent_link, child_link, positionOnly=positionOnly)

    def numpy(self, q: ca.SX, parent_link: str, child_link: str, positionOnly=False):
        fk_parent = super().numpy_by_name(q, parent_link)
        fk_child = super().numpy_by_name(q, child_link)
        tf_parent_child = np.dot(np.linalg.inv(fk_parent), fk_child)
        if positionOnly:
            return tf_parent_child[0:3, 3]
        else:
            return tf_parent_child




from abc import abstractmethod, ABC
import numpy as np
import casadi as ca
import re

from forwardkinematics.fksCommon.fk import ForwardKinematics

class NoLinkIndexFoundInLinkNameError(Exception):
    pass

class ForwardKinematicsPlanar(ForwardKinematics):

    def fk(self, q, link: str, positionOnly: bool=False):
        if isinstance(link, str):
            return super().fk(q, self.get_link_index(link), positionOnly=positionOnly)
        else:
            return super().fk(q, link, positionOnly=positionOnly)

    def get_link_index(self, link: str):
        regex_match = re.match(r'\D*(\d*)', link)
        try:
            return int(regex_match.group(1))
        except Exception as e:
            raise NoLinkIndexFoundInLinkNameError(f"Link name {link} could not be resolved")




from abc import abstractmethod, ABC
import numpy as np
import casadi as ca


class ForwardKinematicsByName(ABC):

    """Abstract class for forward kinematics"""

    def __init__(self):
        ABC.__init__(self)

    def fk(self, q, link: str, positionOnly: bool=False):
        if isinstance(q, ca.SX):
            return self.casadi(q, link, positionOnly=positionOnly)
        elif isinstance(q, np.ndarray):
            return self.numpy(q, link, positionOnly=positionOnly)

    def n(self):
        return self._n

    @abstractmethod
    def casadi(self, q: ca.SX, link: str, positionOnly=False) -> ca.SX:
        pass

    @abstractmethod
    def numpy(self, q: np.ndarray, link: str, positionOnly=False) -> np.ndarray:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['forwardkinematics',
 'forwardkinematics.fksCommon',
 'forwardkinematics.planarFks',
 'forwardkinematics.urdfFks',
 'forwardkinematics.urdfFks.casadiConversion',
 'forwardkinematics.urdfFks.casadiConversion.geometry']

package_data = \
{'': ['*'], 'forwardkinematics.urdfFks': ['urdf/*']}

install_requires = \
['casadi>=3.5.4,<4.0.0,!=3.5.5.post1,!=3.5.5.post2',
 'numpy>=1.15.3,<2.0.0',
 'urdf_parser_py>=0.0.3']

setup_kwargs = {
    'name': 'forwardkinematics',
    'version': '1.0.6',
    'description': '"Light-weight implementation of forward kinematics using casadi."',
    'long_description': '# Installation\n\nThis package provides a forward kinematics for simple robots as symbolic functions using\ncasadi. This allows the usage in model predictive control schemes and other trajectory\noptimization methods.\n\n```bash\npip3 install forwardkinematics\n```\n\n## Install in editable mode\n\nIf you want to install as an editable without the usage of an virtual environment, you\nmust create a setup.py first.\nThis can be done using poetry2setup (`pip install poetry2setup`)\nThen you can run \n```bash\npoetry2setup > setup.py\nmv pyproject.toml pyproject.toml_BACKUP\npip3 install -e .\n```\n',
    'author': 'Max Spahn',
    'author_email': 'm.spahn@tudelft.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/maxspahn/forwardKinematics.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<3.10',
}


setup(**setup_kwargs)



import numpy as np
import casadi as ca
from forwardkinematics.fksCommon.fk import ForwardKinematics


class MobileRobotFk(ForwardKinematics):
    def __init__(self, n, baseHeight=1.0):
        super().__init__()
        self._n = n
        self._baseHeight = baseHeight

    def fk(self, q, i, positionOnly=False, endlink=0.0):
        assert i <= self._n
        if isinstance(q, ca.SX):
            assert q.shape[0] == self._n
            return self.casadi(q, i, positionOnly=positionOnly, endlink=endlink)
        elif isinstance(q, np.ndarray):
            assert q.size == self._n
            return self.numpy(q, i, positionOnly=positionOnly, endlink=endlink)

    def casadi(self, q, i, positionOnly=False, endlink=0.0):
        fk = ca.SX(np.zeros(3))
        if i > 0:
            if self._n > 1:
                fk = ca.SX(np.array([q[0], self._baseHeight + 0.2, q[1]]))
            else:
                fk = ca.SX(np.array([q[0], self._baseHeight + 0.2, 0.0]))
        for i in range(2, i + 1):
            fk[0] += ca.cos(fk[2]) * 1.0
            fk[1] += ca.sin(fk[2]) * 1.0
            if i < q.size(1):
                fk[2] += q[i]
        fk[0] += ca.cos(fk[2]) * endlink
        fk[1] += ca.sin(fk[2]) * endlink
        if positionOnly:
            return fk[0:2]
        else:
            return fk

    def numpy(self, q, i, positionOnly=False, endlink=0.0):
        fk = np.zeros(3)
        if i > 0:
            if self._n > 1:
                fk = np.array([q[0], self._baseHeight + 0.2, q[1]])
            else:
                fk = np.array([q[0], self._baseHeight + 0.2, 0.0])
        for i in range(2, i + 1):
            fk[0] += np.cos(fk[2]) * 1.0
            fk[1] += np.sin(fk[2]) * 1.0
            if i < len(q):
                fk[2] += q[i]
        fk[0] += np.cos(fk[2]) * endlink
        fk[1] += np.sin(fk[2]) * endlink
        if positionOnly:
            return fk[0:2]
        else:
            return fk



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class DualArmFk(URDFForwardKinematics):
    def __init__(self):
        fileName = "dual_arm.urdf"
        relevantLinks = ["world", "link1", "link2", "ee1_link", "link4", "ee2_link"]
        super().__init__(fileName, relevantLinks, "world", ["ee1_link", "ee2_link"], 5)



import numpy as np
import casadi as ca
from forwardkinematics.planarFks.planar_fk import ForwardKinematicsPlanar


class PointFk(ForwardKinematicsPlanar):
    def __init__(self):
        super().__init__()
        self._n = 2

    def fk(self, q, link, positionOnly: bool=False, endlink=0.0):
        if isinstance(link, str):
            i = self.get_link_index(link)
        else:
            i = link
        assert i <= self._n
        if isinstance(q, ca.SX):
            assert q.shape[0] == self._n
            return self.casadi(q, i, positionOnly=positionOnly, endlink=endlink)
        elif isinstance(q, np.ndarray):
            assert q.size == self._n
            return self.numpy(q, i, positionOnly=positionOnly, endlink=endlink)

    def casadi(self, q, i, positionOnly=False, endlink=0.0):
        fk = ca.SX(np.array([0.0, 0.0]))
        if i > 0:
            fk += q
        return fk

    def numpy(self, q, i, positionOnly=False, endlink=0.0):
        fk = np.array([0.0, 0.0])
        if i > 0:
            fk += q
        return fk



import numpy as np
import casadi as ca
from forwardkinematics.fksCommon.fk import ForwardKinematics


class GroundRobotFk(ForwardKinematics):
    def __init__(self, n=3):
        super().__init__()
        self._n = n
        self._ee_link = 0.2

    def fk(self, q, i, positionOnly=False, endlink=0.0):
        assert i <= self._n
        if isinstance(q, ca.SX):
            # assert(q.shape[0] == self._n)
            return self.casadi(q, i, positionOnly=positionOnly, endlink=endlink)
        elif isinstance(q, np.ndarray):
            # assert(q.size == self._n)
            return self.numpy(q, i, positionOnly=positionOnly, endlink=endlink)

    def casadi(self, q, i, positionOnly=False, endlink=0.0):
        fk = ca.SX(np.array([0.0, 0.0, 0.0]))
        if i > 0:
            fk += q[0:3]
        if i > 1:
            fk[0:2] += self._ee_link * ca.vertcat(ca.cos(fk[2]), ca.sin(fk[2]))
        if i > 2 and i < self.n():
            fk[2] += q[i]
            fk[0:2] += ca.vertcat(ca.cos(fk[2]), ca.sin(fk[2]))
        if positionOnly:
            return fk[0:2]
        else:
            return fk

    def numpy(self, q, i, positionOnly=False, endlink=0.0):
        fk = np.array([0.0, 0.0, 0.0])
        if i > 0:
            fk += q[0:3]
            fk[0:2] += self._ee_link * np.array([np.cos(fk[2]), np.sin(fk[2])])
        if i > 2 and i < self.n():
            fk[2] += q[i]
            fk[0:2] += np.array([np.cos(fk[2]), np.sin(fk[2])])
        if positionOnly:
            return fk[0:2]
        else:
            return fk
